/*
   JV Library
   Copyright (C) 2014  Jouko Juhani Virtanen

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   */
# include <algorithm>

using namespace std;

# include "TypeDef.h"
# include "Structures.h"
# include "PrintPdb.h"
# include "AtomUtils.h"
# include "VectorManip.h"
# include "ASA.h"

Real isOverlap(vector<AtomStruct> &Atoms, Real x, Real y, Real z)
{
	int natom=Atoms.size();
	Real dx, dy, dz;
	Real r2, dist, dist2;

	for (int i=0;i<natom;i++)
	{
		dx=Atoms[i].x-x;
		dy=Atoms[i].y-y;
		dz=Atoms[i].z-z;
		r2=dx*dx+dy*dy+dz*dz;
		dist=Atoms[i].vdw+probeRadius;
		dist2=dist*dist;
		if (r2<dist2)
		{
			return 0;
		}
	}
	return probeRadius+1;
}

Real ASAOfAtomSharkeRupley(vector<AtomStruct> &Atoms, int index)
{
	int accessible, points;
	Real probeRadius, d, r, sinphi, cosphi, theta, thetainc, nslices, asa;
	Real VectX, VectY, VectZ;
	Real xgrid, ygrid, zgrid;
	Real overlap;
	accessible=0;
	points=0;
	probeRadius=1.4;
	d=0.0001;                               //This is a hack, because the points should not overlap with the atoms they belong to.
	r=Atoms[index].vdw+probeRadius+d;
	nslices=10;
	thetainc=nslices;
	cosphi=1.0/nslices-1.0;
	theta=pi/thetainc;
	while (true)
	{
		sinphi=sqrt(1.0-cosphi*cosphi);
		VectX=sinphi*cos(theta);
		VectY=sinphi*sin(theta);
		VectZ=cosphi;
		theta+=2.0*pi/thetainc;
		if (theta>2.0*pi)
		{
			theta=pi/thetainc;
			cosphi+=2.0/nslices;
			if (cosphi>1.0) break;
		}
		//cout <<"theta= "<<theta<<" cosphi= "<<cosphi<<endl;
		xgrid=Atoms[index].x+VectX*r;
		ygrid=Atoms[index].y+VectY*r;
		zgrid=Atoms[index].z+VectZ*r;
		overlap=isOverlap(Atoms, xgrid, ygrid, zgrid);
		//cout <<"overlap= "<<overlap<<endl;
		if (overlap>probeRadius) accessible++;
		points++;
	}
	//cout <<"accessible= "<<accessible<<" points= "<<points<<endl;
	asa=Real(accessible)*4.0*pi*r*r/Real(points);
	return asa;
}

Real ASASharkeRupley(vector<AtomStruct> &Atoms)
{
	int natom;
	Real asa=0;

	natom=findNumProteinAtoms(Atoms);
	for (int i=0;i<natom;i++)
	{
		asa+=ASAOfAtomSharkeRupley(Atoms, i);
	}
	return asa;
}

void ASASharkeRupley(vector<AtomStruct> &Atoms, Vector &asas)
{
	int natom;

	natom=findNumProteinAtoms(Atoms);
	SafeAlloc(asas, natom, "asas");
	for (int i=0;i<natom;i++)
	{
		asas[i]=ASAOfAtomSharkeRupley(Atoms, i);
	}
}

Real FindSAOverlap(AtomStruct &Atom1, AtomStruct &Atom2, Real dist)
{
	Real r1, r2, costheta, overlap;
	r1=Atom1.vdw+probeRadius;
	r2=Atom2.vdw+probeRadius;
	if (r1+r2<dist) return 0;
	if (r2>dist+r1) return 0;
	if (r1>dist+r2) return 4.0*pi*r2*r2;
	if (dist==0) return 4.0*pi*r2*r2;
	costheta=(r2*r2+dist*dist-r1*r1)/(2.0*dist*r2);
	overlap=2.0*pi*r2*r2*(1.0-costheta);
	if (isnan(overlap))
	{
		cout <<"r1= "<<r1<<" r2= "<<r2<<" dist= "<<dist<<" costheta= "<<costheta<<" overlap= "<<overlap<<endl;
	}
	return overlap;
}

Real ASAOfAtomLCPO(vector<AtomStruct> &Atoms, int index, Matrix &DistMatrix)
{
	//Need to read actual paper
	int natom;
	Real overlap, r, asa, area;
	natom=findNumProteinAtoms(Atoms);
	overlap=0;
	r=Atoms[index].vdw+probeRadius;
	area=4.0*pi*r*r;
	asa=area;
	if (r!=0)
	{
		for (int i=0;i<natom;i++)
		{
			if (i!=index)
			{
				overlap=FindSAOverlap(Atoms[i], Atoms[index], DistMatrix[i][index]);
				if (isnan(overlap))
				{
					printAtomInfo(Atoms[i]);
					printAtomInfo(Atoms[index]);
				}
				//cout <<"asa= "<<asa<<" overlap= "<<overlap<<" area= "<<area<<endl;
				asa*=(area-overlap)/area;
			}
		}
	}
	//endProgram(__LINE__, __FILE__);
	return asa;
}

Real ASALCPO(vector<AtomStruct> &Atoms, Matrix &DistMatrix)
{
	int natom;
	Real asa=0;

	natom=findNumProteinAtoms(Atoms);
	for (int i=0;i<natom;i++)
	{
		asa+=ASAOfAtomLCPO(Atoms, i, DistMatrix);
	}
	return asa;
}

Real ASALCPO(vector<AtomStruct> &Atoms)
{
	Real asa=0;
	Matrix DistMatrix;

	calcDistanceMatrix(Atoms, DistMatrix);
	asa=ASALCPO(Atoms, DistMatrix);

	return asa;
}

void ASALCPO(vector<AtomStruct> &Atoms, Matrix &DistMatrix, Vector &asas)
{
	int natom;
	Real asa=0;

	natom=findNumProteinAtoms(Atoms);
	for (int i=0;i<natom;i++)
	{
		asa=ASAOfAtomLCPO(Atoms, i, DistMatrix);
		SafePushBack(asas, asa, "asas");
	}
}

void calcPercentAccessibleByResidue(vector<AtomStruct> &Atoms, Vector &asas)
{
	int natom=Atoms.size();
	Real r, totalAsa=0, maxPossible=0;
	//Real percent;

	for (int i=0;i<natom;i++)
		//for (int i=0;i<5;i++)
	{
		r=Atoms[i].vdw+probeRadius;
		totalAsa+=asas[i];
		maxPossible+=4.0*pi*r*r;
		//cout <<"i= "<<i<<" asa= "<<asas[i]<<endl;
		if (i==natom-1 || Atoms[i].ResidueNum!=Atoms[i+1].ResidueNum)
		{
			//percent=totalAsa*100.0/maxPossible;
			//cout <<Atoms[i].ResidueNum<<"\t"<<Atoms[i].ResidueName<<"\t"<<totalAsa<<"\t"<<maxPossible<<"\t"<<percent<<endl;
			cout <<totalAsa<<endl;
			totalAsa=0;
			maxPossible=0;
		}
	}
}

void calcPercentAccessibleByResidue(vector<AtomStruct> &Atoms)
{
	Vector asas;
	Matrix DistMatrix;

	calcDistanceMatrix(Atoms, DistMatrix);
	ASALCPO(Atoms, DistMatrix, asas);
	calcPercentAccessibleByResidue(Atoms, asas);
}

void calcPercentAccessibleByResidueSharke(vector<AtomStruct> &Atoms)
{
	Vector asas;
	Matrix DistMatrix;

	ASASharkeRupley(Atoms, asas);
	calcPercentAccessibleByResidue(Atoms, asas);
}

/*
   void QuickSurfaceASA(vector<AtomStruct> &Atoms, vector<AtomStruct> &SurfaceAtoms)
   {
   int natom;
   Real threshold, asa;
   natom=findNumProteinAtoms(Atoms);
   threshold=1.0;
   for (int i=0;i<natom;i++)
   {
   asa=ASAOfAtomLCPO(Atoms, i);
   cout <<"asa= "<<asa<<endl;
   if (asa>threshold) SafePushBack(SurfaceAtoms, Atoms[i], "SurfaceAtoms");
   }
   }

   void QuickSurface(vector<AtomStruct> Atoms, vector<AtomStruct> &Surface)
   {
   vector<bool> BoolSurface;
   int natom, CurrentAtom;
   Real sinphi, cosphi, theta;
   Real VectX, VectY, VectZ;
   Real thetainc, nslices=100;
   Real current, projection;
   Real xn, yn, zn;
   center(Atoms);

   thetainc=nslices;
   cosphi=1.0/nslices-1.0;
   theta=pi/thetainc;
   natom=findNumProteinAtoms(Atoms);
   SafeAlloc(BoolSurface, false, natom, "BoolSurface");

   while (true)
   {
   sinphi=sqrt(1.0-cosphi*cosphi);
   VectX=sinphi*cos(theta);
   VectY=sinphi*sin(theta);
   VectZ=cosphi;
   theta+=2.0*pi/thetainc;
   if (theta>2.0*pi)
   {
   theta=pi/thetainc;
   cosphi+=2.0/nslices;
   if (cosphi>1.0) break;
   }
   cout <<"theta= "<<theta<<" cosphi= "<<cosphi<<endl;
   current=0;
   for (int n=0;n<natom;n++)
   {
   xn=Atoms[n].x;
   yn=Atoms[n].y;
   zn=Atoms[n].z;
   projection=VectX*xn+VectY*yn+VectZ*zn;
   if (projection+Atoms[n].HydrationRadius>current)
   {
   current=projection+Atoms[n].HydrationRadius;
   CurrentAtom=n;
   }
   }
   BoolSurface[CurrentAtom]=true;
   }
   for (int i=0;i<natom;i++)
   {
   if (BoolSurface[i]) SafePushBack(Surface, Atoms[i], "Surface");
   }
   }
   */

