# include <iostream>
# include <vector>
# include <algorithm>

using namespace std;

# include "TypeDef.h"
# include "Structures.h"
# include "AtomUtils.h"
# include "VectorManip.h"

Real calcHBondEnergy(ProteinStruct &Protein, int res1, int res2)
{
	int co, o, n, hn;
	Real ron, rch, roh, rcn;
	Real hBondEnergy;

	co=getAtomIndex(Protein.Atoms, res1, "C");
	o=getAtomIndex(Protein.Atoms, res1, "O");
	n=getAtomIndex(Protein.Atoms, res2, "N");
	hn=getAtomIndex(Protein.Atoms, res2, "HN");

	if (co==UNK_INT || o==UNK_INT || n==UNK_INT || hn==UNK_INT)
	{
		return 3.0;
	}

	ron=atomDistance(Protein.Atoms[o], Protein.Atoms[n]);
	rch=atomDistance(Protein.Atoms[co], Protein.Atoms[hn]);
	roh=atomDistance(Protein.Atoms[o], Protein.Atoms[hn]);
	rcn=atomDistance(Protein.Atoms[co], Protein.Atoms[n]);

	cout <<"ron= "<<ron<<" rch= "<<rch<<" roh= "<<roh<<" rcn= "<<rcn<<" res1= "<<res1<<" res2= "<<res2<<endl;

	hBondEnergy=1.0/ron+1.0/rch-1.0/roh-1.0/rcn;
	hBondEnergy*=0.42*0.2*332.0; //kcal/mol

	if (res2-res1==4)
	{
		cout <<"hBondEnergy= "<<hBondEnergy<<endl;
	}
	return hBondEnergy;	
}

bool isHBond(ProteinStruct &Protein, int res1, int res2, Real cutoff)
{
	Real hBondEnergy;

	hBondEnergy=calcHBondEnergy(Protein, res1, res2);
	if (hBondEnergy<cutoff) return true;
	return false;
}

void getNhoBonding(ProteinStruct &Protein, int res, int &nho1, int &nho2, Real &nho1Energy, Real &nho2Energy)
{
	int firstRes, lastRes;
	Real energy;

	nho1=nho2=0;
	nho1Energy=nho2Energy=0;
	getFirstResidue(Protein.Atoms, firstRes);
	getLastResidue(Protein.Atoms, lastRes);

	cout <<endl;
	for (int i=firstRes;i<=lastRes;i++)
	{
		if (res!=i && res!=i+1)
		{
			energy=calcHBondEnergy(Protein, i, res);
			if (energy<nho1Energy || i==firstRes)
			{
				nho2=nho1;
				nho2Energy=nho1Energy;
				nho1=i;
				nho1Energy=energy;
			}
			else if (energy<nho2Energy)
			{
				nho2=i;
				nho2Energy=energy;
			}
			cout <<"res= "<<res<<" i= "<<i<<" energy= "<<energy<<" nho1Energy= "<<nho1Energy<<" nho1= "<<nho1<<endl;
		}
	}
	cout <<endl;
}

void getOhnBonding(ProteinStruct &Protein, int res, int &ohn1, int &ohn2, Real &ohn1Energy, Real &ohn2Energy)
{
	int firstRes, lastRes;
	Real energy;

	ohn1=ohn2=0;
	ohn1Energy=ohn2Energy=0;
	getFirstResidue(Protein.Atoms, firstRes);
	getLastResidue(Protein.Atoms, lastRes);

	for (int i=firstRes;i<=lastRes;i++)
	{
		if (res!=i && res!=i-1)
		{
			energy=calcHBondEnergy(Protein, res, i);
			if (energy<ohn1Energy || i==firstRes)
			{
				ohn2=ohn1;
				ohn2Energy=ohn1Energy;
				ohn1=i;
				ohn1Energy=energy;
			}
			else if (energy<ohn2Energy)
			{
				ohn2=i;
				ohn2Energy=energy;
			}
		}
	}
}

void getHydrogenBonding(ProteinStruct &Protein, int res, int &nho1, int &ohn1, int &nho2, int &ohn2, Real &nho1Energy, Real &ohn1Energy, Real &nho2Energy, Real &ohn2Energy)
{
	getNhoBonding(Protein, res, nho1, nho2, nho1Energy, nho2Energy);
	getOhnBonding(Protein, res, ohn1, ohn2, ohn1Energy, ohn2Energy);
	cout <<"res= "<<res<<" nho1= "<<nho1<<endl;
}

void findHydrogenBondingNetwork(ProteinStruct &Protein, HydrogenBondingStruct &hBonding)
{
	int firstRes, lastRes;
	int nho1, ohn1, nho2, ohn2;
	Real nho1Energy, ohn1Energy, nho2Energy, ohn2Energy;

	getFirstResidue(Protein.Atoms, firstRes);
	getLastResidue(Protein.Atoms, lastRes);

	for (int i=firstRes;i<=lastRes;i++)
	{
		getHydrogenBonding(Protein, i, nho1, ohn1, nho2, ohn2, nho1Energy, ohn1Energy, nho2Energy, ohn2Energy);
		SafePushBack(hBonding.nho1, nho1, "nho1");
		SafePushBack(hBonding.ohn1, ohn1, "ohn1");
		SafePushBack(hBonding.nho2, nho2, "nho2");
		SafePushBack(hBonding.ohn2, ohn2, "ohn2");
		SafePushBack(hBonding.nho1Energy, nho1Energy, "nho1Energy");
		SafePushBack(hBonding.ohn1Energy, ohn1Energy, "ohn1Energy");
		SafePushBack(hBonding.nho2Energy, nho2Energy, "nho2Energy");
		SafePushBack(hBonding.ohn2Energy, ohn2Energy, "ohn2Energy");
	}
}

void findHBonds(ProteinStruct &Protein, vector< vector<int> > &hbonds, Real cutoff)
{
	bool hbond;
	int firstRes=UNK_INT, lastRes=UNK_INT;
	int nres;

	getFirstResidue(Protein.Atoms, firstRes);
	getLastResidue(Protein.Atoms, lastRes);
	nres=lastRes-firstRes+1;

	Safe2DAlloc(hbonds, nres, 0, "hbonds");
	for (int i=firstRes;i<=lastRes;i++)
	{
		for (int j=i+1;j<=lastRes;j++)
		{
			hbond=isHBond(Protein, i, j, cutoff);
			if (hbond)
			{
				SafePushBack(hbonds[i-firstRes], j-firstRes, "hbonds");
			}
		}
	}
}

void findHBonds(ProteinStruct &Protein, vector< vector<int> > &hbonds)
{
	Real cutoff=-0.5;

	findHBonds(Protein, hbonds, cutoff);
}
