/*
    JV Library
    Copyright (C) 2014  Jouko Juhani Virtanen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef _MonteCarloTemplate_included_
#define _MonteCarloTemplate_included_

# include <sys/time.h>

# include "MathUtils.h"
# include "VectorManip.h"
# include "MinMax.h"
# include "MonteCarlo.h"
# include "time.h"

        template<class Args>
Real monteCarlo(Vector &coefficient, Args &args, Real (*func_to_minimize)(Vector&, Args&), Real temp, int nsteps, MoveStruct &move)
{
	int updateMoveSizeInterval=50;
	int nMoveTypes=coefficient.size(), nmove;
	Real score, score_old, score_initial, dScore, randNum;
	Real acceptProb, acceptRate;
	Vector old_coefficient;

	score_initial=(*func_to_minimize)(coefficient, args);
	score_old=score_initial;
	old_coefficient=coefficient;
	for (int i=0;i<nsteps;i++)
	{
		makeRandomChange(coefficient, move);
		score=(*func_to_minimize)(coefficient, args);
		dScore=score-score_old;
		acceptProb=calcAcceptProb(dScore, temp);
		SafePushBack(move.score, score, "score");
		SafePushBack(move.dScore, dScore, "dScore");
		SafePushBack(move.temp, temp, "temp");
		SafePushBack(move.acceptProb, acceptProb, "acceptProb");
		randNum=randDouble(0,1.0);
		if (acceptProb>randNum)
		{
			score_old=score;
			old_coefficient=coefficient;
		}
		else
		{
			coefficient=old_coefficient;
		}
		nmove=move.score.size();
		if (nmove%updateMoveSizeInterval==0 && nmove!=0) 
		{
			updateMoveSizes(move, nMoveTypes);
		}
	}
	acceptRate=calcAcceptRate(move);
	cout <<"score_initial= "<<score_initial<<" score_final= "<<score_old
	<<" temp= "<<temp<<" acceptRate= "<<acceptRate<<endl;
	return score_old;
}

	template<class Args>
Vector calcScores(Matrix &coefficients, Args &args, Real (*func_to_minimize)(Vector&, Args&))
{
	int nrep=coefficients.size();
	Vector scores;

	SafeAlloc(scores, nrep, "scores");
	for (int i=0;i<nrep;i++)
	{
		scores[i]=func_to_minimize(coefficients[i], args);
	}
	return scores;
}


	template<class Args>
void multipleSwaps(Matrix &coefficients, Args &args, Real (*func_to_minimize)(Vector&, Args&), Vector &totalSwaps, Vector &swapAttempts, Vector &temps, vector<int> &hotOrCold, Vector &hotHistogram, Vector &coldHistogram)
{
	Vector scores;

	scores=calcScores(coefficients, args, func_to_minimize);
	multipleSwaps(coefficients, scores, totalSwaps, swapAttempts, temps, hotOrCold, hotHistogram, coldHistogram);
}

	template<class Args>
Real replicaExchange(Matrix &coefficients, Vector &bestCoefficient, Args &args, Real (*func_to_minimize)(Vector&, Args&), Real minTemp, Real maxTemp, int nrep, int nsteps, int ncycles, Real maxHours)
{
	int best, Size;
	vector<int> hotOrCold;
	Real bestScore, finalScore, hours;
	Vector score, temps, totalSwaps, swapAttempts;
	Vector hotHistogram, coldHistogram;
	vector<MoveStruct> moves;
	timeval start, end;
	
	gettimeofday(&start, NULL);
	printMatrix(coefficients);
	
	SafeAlloc(score, nrep, "score");
	SafeAlloc(totalSwaps, nrep, "totalSwaps");
	SafeAlloc(swapAttempts, nrep, "swapAttempts");
	setTemps(temps, minTemp, maxTemp, nrep);
	initializeHotAndCold(hotOrCold, hotHistogram, coldHistogram, nrep);
	Size=coefficients[0].size();
	initializeMoves(coefficients, moves);
	multipleSwaps(coefficients, args, func_to_minimize, totalSwaps, swapAttempts, temps, hotOrCold, hotHistogram, coldHistogram);
	zeroVector(totalSwaps); zeroVector(swapAttempts);
	bestScore=func_to_minimize(coefficients[0], args);
	bestCoefficient=coefficients[0];

	for (int i=0;i<ncycles;i++)
	{
		for (int j=0;j<nrep;j++)
		{
			score[j]=monteCarlo(coefficients[j], args, func_to_minimize, temps[j], nsteps, moves[j]);
			if (score[j]<bestScore)
			{
				bestScore=score[j];
				bestCoefficient=coefficients[j];
			}
		}
		multipleSwaps(coefficients, args, func_to_minimize, totalSwaps, swapAttempts, temps, hotOrCold, hotHistogram, coldHistogram);
		best=findMin(score, finalScore);
		for (int k=0;k<Size;k++)
		{
			cout <<"coefficient["<<k<<"]= "<<coefficients[best][k]<<endl;
		}
		gettimeofday(&end, NULL);
		hours=calcTimeDiff(start, end)/3600.0;
		if (hours>maxHours) break;
	}

	printSwapRates(totalSwaps, swapAttempts);
	monteCarlo(coefficients[0], args, func_to_minimize, temps[0], nsteps*2, moves[0]); //Temporary. Delete line.
	for (int i=0;i<Size;i++)
	{
		cout <<"coefficient["<<i<<"]= "<<coefficients[0][i]<<endl;
	}
	cout <<"bestScore= "<<bestScore<<endl;	
	return bestScore;
}
	template<class Args>
Real replicaExchange(Matrix &coefficients, Vector &bestCoefficient, Args &args, Real (*func_to_minimize)(Vector&, Args&), Real minTemp, Real maxTemp, int nrep, int nsteps, int ncycles)
{
	int best, Size;
	vector<int> hotOrCold;
	Real bestScore, finalScore;
	Vector score, temps, totalSwaps, swapAttempts;
	Vector hotHistogram, coldHistogram;
	vector<MoveStruct> moves;
	printMatrix(coefficients);
	
	SafeAlloc(score, nrep, "score");
	SafeAlloc(totalSwaps, nrep, "totalSwaps");
	SafeAlloc(swapAttempts, nrep, "swapAttempts");
	setTemps(temps, minTemp, maxTemp, nrep);
	initializeHotAndCold(hotOrCold, hotHistogram, coldHistogram, nrep);
	Size=coefficients[0].size();
	initializeMoves(coefficients, moves);
	multipleSwaps(coefficients, args, func_to_minimize, totalSwaps, swapAttempts, temps, hotOrCold, hotHistogram, coldHistogram);
	zeroVector(totalSwaps); zeroVector(swapAttempts);
	bestScore=func_to_minimize(coefficients[0], args);
	bestCoefficient=coefficients[0];

	for (int i=0;i<ncycles;i++)
	{
		for (int j=0;j<nrep;j++)
		{
			score[j]=monteCarlo(coefficients[j], args, func_to_minimize, temps[j], nsteps, moves[j]);
			if (score[j]<bestScore)
			{
				bestScore=score[j];
				bestCoefficient=coefficients[j];
			}
		}
		multipleSwaps(coefficients, args, func_to_minimize, totalSwaps, swapAttempts, temps, hotOrCold, hotHistogram, coldHistogram);
		best=findMin(score, finalScore);
		for (int k=0;k<Size;k++)
		{
			cout <<"coefficient["<<k<<"]= "<<coefficients[best][k]<<endl;
		}
	}

	printSwapRates(totalSwaps, swapAttempts);
	monteCarlo(coefficients[0], args, func_to_minimize, temps[0], nsteps*2, moves[0]); //Temporary. Delete line.
	for (int i=0;i<Size;i++)
	{
		cout <<"coefficient["<<i<<"]= "<<coefficients[0][i]<<endl;
	}
	cout <<"bestScore= "<<bestScore<<endl;	
	return bestScore;
}

	template<class Args>
Real replicaExchange(Vector &coefficient, Vector &bestCoefficient, Args &args, Real (*func_to_minimize)(Vector&, Args&), Real minTemp, Real maxTemp, int nrep, int nsteps, int ncycles)
{
	Real score;
	Matrix coefficients;

	initializeCoefficients(coefficient, coefficients, nrep);
	score=replicaExchange(coefficients, bestCoefficient, args, func_to_minimize, minTemp, maxTemp, nrep, nsteps, ncycles);
	coefficient=coefficients[0];
	return score;
}

#endif
